@@ -24,7 +24,7 @@
 # OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 # SUCH DAMAGE.
 #
-# $Id: pkg_replace.sh,v 1.13 2006/10/25 13:01:06 securedog Exp $
+# $Id: pkg_replace.sh,v 1.15 2006/10/31 14:37:39 securedog Exp $
 
 usage() {
 	echo "usage: ${0##*/} [-habBcCiInqrRuvVwW] [-l file] [-m make_args]"
@@ -88,6 +88,7 @@ init_options() {
 	BEFOREDEINSTALL=
 	AFTERINSTALL=
 	IGNORE=
+	USE_PKGS=
 #endif
 }
 
@@ -270,8 +271,9 @@ parse_exec_script() {
 set_port_name() {
 	port_name=
 	if is_yes ${opt_use_index}; then
+		# XXX: requires GNU grep.
 		port_name=$(grep --mmap -m 1 -o "^[^|]*|[^|]*${1#${PORTSDIR}/}|" "${PORTSDIR}/INDEX")
-		port_name=${port_name%%|*}
+		port_name=${port_name%%\|*}
 	elif [ -e "$1/Makefile" ] && cd "$1"; then
 #ifdef WITH_PKGSRC
 		port_name=$(${MAKE_CMD} -V '${PKGNAME}')
@@ -298,7 +300,11 @@ set_port_name_from_package() {
 	*.bz2|*.tbz)	opt=j ;;
 	esac
 
+#ifdef WITH_OPENBSD
 	port_name=`tar x${opt}Of "$1" "+CONTENTS" |
+#else
+	port_name=`tar x${opt}Of "$1" --fast-read "+CONTENTS" |
+#endif
 	while read LINE; do
 		case ${LINE} in
 		@name\ *)
@@ -309,7 +315,7 @@ set_port_name_from_package() {
 	done`
 
 	if [ -z "${port_name}" ]; then
-		warn "Unable to open contents file: ${file} (not a package?)"
+		warn "Unable to open contents file: $1 (not a package?)"
 		return 1
 	fi
 }
@@ -370,7 +376,6 @@ correct_pkg() {
 		fi
 		return 1
 	elif [ ! -d "${pkg_portdir}" ]; then
-		warn "${pkg_name} (${pkg_origin}): perhaps moved or obsoleted."
 		trace_moved "${pkg_origin}" || return 1
 	fi
 }
@@ -540,7 +545,7 @@ try() {
 	local exit
 
 	if [ ${opt_debug} -ge 2 ]; then
-		echo "[$@]"
+		warn "$@"
 	elif "$@"; exit=$?; [ ${exit} != 0 ]; then
 		warn "Command failed (exit code ${exit}): $@"
 		return ${exit}
@@ -548,7 +553,16 @@ try() {
 }
 
 build_port() {
-	info "Building '$1'"
+	local msg
+
+#ifdef WITH_OPENBSD
+	msg="${MAKE_CMD#env }"
+	msg="${MAKE_CMD% ${MAKE}}"
+#else
+	msg="${MAKE_CMD#${MAKE} }"
+#endif
+
+	info "Building '$1'${msg:+ with make flags:${msg}}"
 	cd "$1" || return 1
 
 #ifdef WITH_CONFIG
@@ -637,6 +651,80 @@ clean_port() {
 	try ${MAKE_CMD} ${clean_args} clean || return 1
 }
 
+#ifdef WITH_CONFIG
+fetch_package() {
+	local pkg_uri prog subdir OPSYS OS_VERSION OS_REVISION OS_MAJOR ARCH
+
+	info "Fetching the binary package for '$1'"
+
+	set -- "$1" "${2:-${cwd}}" ${UNAME=$(uname -smr)}
+
+	OPSYS=$3
+	OS_VERSION=$4
+	OS_REVISION=${OS_VERSION%%-*}
+	OS_MAJOR=${OS_REVISION%%.*}
+	ARCH=$5
+
+	if [ ! -w "$2" ]; then
+		warn "You do not own $2."
+		return 1
+	fi
+
+	if [ ${PACKAGESITE:+1} ]; then
+		pkg_uri=${PACKAGESITE}
+	else
+#ifdef WITH_PKGSRC
+		pkg_uri="${PACKAGEROOT:-ftp://ftp.NetBSD.org}/pub/pkgsrc/packages/${OPSYS}-${OS_REVISION}/${ARCH}/All/"
+#else
+#ifdef WITH_OPENBSD
+		pkg_uri="${PACKAGEROOT:-ftp://ftp.OpenBSD.org}/pub/OpenBSD/${OS_REVISION}/packages/${ARCH}/"
+#else
+		case ${OS_VERSION} in
+		*-CURRENT)	subdir="${OS_MAJOR}-current" ;;
+		*-RELEASE*)	subdir="${OS_REVISION}-release" ;;
+		*)		subdir="${OS_MAJOR}-stable" ;;
+		esac
+
+		pkg_uri="${PACKAGEROOT:-ftp://ftp.FreeBSD.org}/pub/FreeBSD/ports/${ARCH}/packages-${subdir}/All/"
+#endif
+#endif
+	fi
+	pkg_uri="${pkg_uri}$1${PKG_SUFX}"
+
+	cd "$2" || return 1
+
+	for prog in "/usr/bin/fetch -Ap" /usr/bin/ftp /usr/local/bin/ftp; do
+		[ -x "${prog%% *}" ] || continue
+
+		if try ${prog} "${pkg_uri}" && [ -e "$1${PKG_SUFX}" ]; then
+			return 0
+		else
+			return 1
+		fi
+	done
+}
+
+find_package() {
+	use_package="${PKGREPOSITORY}/${1}${PKG_SUFX}"
+
+	if ! {
+		[ -e "${use_package}" ] || \
+		fetch_package "$1" "${PKGREPOSITORY}"
+	}; then
+		warn "Couldn't find the latest binary package."
+
+		if is_yes ${opt_force}; then
+			use_package=
+			warn "Using the source files instead of a binary package."
+		else
+			warn "Please put a binary package (${use_package##*/})" \
+			"in '${PKGREPOSITORY}'. (or specify -f to force)"
+			return 1
+		fi
+	fi
+}
+#endif
+
 #ifdef WITH_PKGSRC
 pkg_backup() {
 	local pkg_pkgdir package plist create_args arg
@@ -680,8 +768,8 @@ pkg_backup() {
 }
 #endif
 
-create_old_package() {
-	if is_yes ${opt_backup}; then
+backup_package() {
+	if is_yes ${opt_backup} && [ ! -e "$2" ]; then
 		info "Backing up the old version"
 #ifdef WITH_PKGSRC
 		try pkg_backup "$1" "$2" || return 1
@@ -691,7 +779,7 @@ create_old_package() {
 			cd "${2%/*}" && \
 			try pkg_create -f "${PKG_DBDIR}/$1/+CONTENTS" && \
 			cd "${cwd}"
-		}: then
+		}; then
 			return 1
 		fi
 #else
@@ -701,9 +789,9 @@ create_old_package() {
 	fi
 }
 
-create_old_required_by() {
+backup_file() {
 	if [ -e "$1" ]; then
-		info "Backing up the +REQUIRED_BY file"
+		info "Backing up the ${1##*/} file"
 		try cp -f "$1" "$2" || return 1
 	fi
 }
@@ -715,14 +803,14 @@ restore_package() {
 	fi
 }
 
-restore_required_by() {
+restore_file() {
 	if [ -e "$1" ] && [ ! -e "$2" ]; then
-		info "Restoring the +REQUIRED_BY file"
+		info "Restoring the ${1##*/} file"
 		try mv -f "$1" "$2" || return 1
 	fi
 }
 
-process_old_package() {
+process_package() {
 	if is_yes ${opt_keep_backup} && [ -e "$1" ] && [ ! -e "${PKG_BACKUP_DIR}/${1##*/}" ]; then
 		info "Keeping old package in '${PKG_BACKUP_DIR}'"
 		make_dirs "${PKG_BACKUP_DIR}" || return 1
@@ -736,10 +824,12 @@ cleanup() {
 
 trace_moved() {
 #ifdef WITH_PKGSRC
+	warn "'$1' has been moved or obsoleted."
 	warn "See ${PORTSDIR}/doc/CHANGES-*."
 	return 1
 #else
 #ifdef WITH_OPENBSD
+	warn "'$1' has been moved or obsoleted."
 	return 1
 #else
 	local IFS msg checked moved origin
@@ -751,18 +841,20 @@ trace_moved() {
 	checked=
 	IFS='|'
 
-	while moved=$(grep -m1 "^$1|" "${PORTSDIR}/MOVED"); do
+	while moved=$(grep "^$1|" "${PORTSDIR}/MOVED"); do
 		set -- ${moved}
 
-		if [ -n "$2" ]; then
-			checked="${checked:-|}$1|"
-
+		if [ $# -ne 4 ]; then
+			warn "'${PORTSDIR}/MOVED' has broken!"
+			break
+		elif [ -n "$2" ]; then
 			case ${checked} in
-			*"|${2}|"*)
-				warn "Failed. (${msg} -> ???)"
+			*"|$2|"*)
+				warn "'${PORTSDIR}/MOVED' has broken!"
 				break ;;
 			esac
 
+			checked="${checked:-|}$1|"
 			msg="${msg:+${msg} }-> $2"
 			set_pkg_portdir "$2"
 
@@ -780,7 +872,7 @@ trace_moved() {
 		fi
 	done
 
-	warn "Port directory not found: ${origin}"
+	warn "'$1' has been moved or obsoleted."
 	return 1
 #endif
 #endif
@@ -792,11 +884,8 @@ perform() {
 	while [ $# -gt 0 ]; do
 		log=
 		status=
-		set_pkg_name "$1"
-		set_pkg_pkgdir "$1"
-		shift
 
-		if perform_preprocess; then
+		if perform_preprocess "$1"; then
 			if is_yes ${opt_version}; then
 				perform_version
 			elif ! perform_replace; then
@@ -806,16 +895,19 @@ perform() {
 			fi
 		else
 			status=skipped
-			warn "Skipping '${pkg_name}'."
+			warn "Skipping '$1'."
 		fi
 
-		add_log "${status:-ignored}" "${pkg_name}" ${log:+" (${log})"}
+		add_log "${status:-ignored}" "$1" ${log:+" (${log})"}
+		shift
 	done
 }
 
 perform_preprocess() {
 	local IFS i replace_with
 
+	set_pkg_name "$1"
+	set_pkg_pkgdir "$1"
 	use_package=
 	replace_with=
 
@@ -827,10 +919,11 @@ perform_preprocess() {
 		esac
 	done
 
-	set_pkg_origin "${pkg_name}" 
 	case ${replace_with} in
-	*.t[bg]z)	use_package=${replace_with} ;;
-	?*)		pkg_origin=${replace_with#${PORTSDIR}/} ;;
+	*.t[bg]z|'')
+		set_pkg_origin "${pkg_name}" 
+		[ -z "${replace_with}" ] || use_package=${replace_with} ;;
+	*)	pkg_origin=${replace_with#${PORTSDIR}/} ;;
 	esac
 
 #ifdef WITH_CONFIG
@@ -854,14 +947,16 @@ perform_preprocess() {
 #else
 		MAKE_CMD="${MAKE} ${opt_make_args}"
 #endif
+#ifdef WITH_CONFIG
 		IFS='
 '
-		for i in ${MAKE_ARGS:+${MAKE_ARGS}}; do
+		for i in ${MAKE_ARGS}; do
 			if pkg_match "${i%%=*}"; then
 				MAKE_CMD="${MAKE_CMD} ${i#*=}"
 			fi
 		done
 		IFS=' '
+#endif
 #ifdef WITH_OPENBSD
 		MAKE_CMD="${MAKE_CMD} ${MAKE}"
 #endif
@@ -872,14 +967,15 @@ perform_preprocess() {
 }
 
 perform_replace() {
-	local REPLACE_TMPDIR old_package old_required_by preserved_files
+	local tmpdir old_package old_required_by preserved_files
 
 	if [ "${pkg_name}" != "${port_name}" ]; then
 		info "Replacing '${pkg_name}' with '${port_name}'"
 	elif is_yes ${opt_force}; then
 		info "Reinstalling '${pkg_name}'"
 	else
-		is_yes ${opt_verbose} && warn "No need to replace '${pkg_name}'. (specify -f to force)"
+		is_yes ${opt_verbose} && \
+		warn "No need to replace '${pkg_name}'. (specify -f to force)"
 		return 0
 	fi
 
@@ -890,31 +986,46 @@ perform_replace() {
 		prompt_yesno || return 0
 	fi
 
+#ifdef WITH_CONFIG
+	if [ -z "${use_package}" ] && pkg_match ${USE_PKGS}; then
+		find_package "${port_name}" || { log="package not found"; return 1; }
+	fi
+
+#endif
 	if [ -z "${use_package}" ]; then
 		build_port "${pkg_portdir}" || { log="build error"; return 1; }
 	fi
 
-	REPLACE_TMPDIR=${PKG_TMPDIR}/${pkg_name}.XXXXXX
-
-	if [ ${opt_debug} -lt 2 ]; then
-		REPLACE_TMPDIR=$(mktemp -d "${REPLACE_TMPDIR}") || return 1
+	if [ ${opt_debug} -ge 2 ]; then
+		tmpdir="${PKG_TMPDIR}/${pkg_name}.XXXXXX"
+	else
+		tmpdir=$(mktemp -d "${PKG_TMPDIR}/${pkg_name}.XXXXXX") || {
+			log="backup error"; return 1; }
 	fi
 
+	old_package="${pkg_name}${PKG_SUFX}"
+
+	if [ -e "${PKG_BACKUP_DIR}/${old_package}" ]; then
+		info "Found a package of the old version"
+		old_package="${PKG_BACKUP_DIR}/${old_package}"
+	else
 #ifdef WITH_OPENBSD
-	# PKG_SUFX doesn't work. See src/usr.sbin/pkg_add/pkg_create.
-	old_package="${REPLACE_TMPDIR}/${pkg_name}.tgz"
+		# PKG_SUFX doesn't work, see src/usr.sbin/pkg_add/pkg_create.
+		old_package="${tmpdir}/${pkg_name}.tgz"
 #else
-	old_package="${REPLACE_TMPDIR}/${pkg_name}${PKG_SUFX}"
+		old_package="${tmpdir}/${old_package}"
 #endif
-	old_required_by="${REPLACE_TMPDIR}/+REQUIRED_BY"
+	fi
+
+	old_required_by="${tmpdir}/+REQUIRED_BY"
 
 	if ! {
-		create_old_package "${pkg_name}" "${old_package}" && \
-		create_old_required_by "${pkg_pkgdir}/+REQUIRED_BY" "${old_required_by}" && \
+		backup_package "${pkg_name}" "${old_package}" && \
+		backup_file "${pkg_pkgdir}/+REQUIRED_BY" "${old_required_by}" && \
 		preserve_shlibs "${pkg_name}"
 	}; then
 		log="backup error"
-		cleanup "${REPLACE_TMPDIR}"
+		cleanup "${tmpdir}"
 		return 1
 	fi
 
@@ -927,30 +1038,36 @@ perform_replace() {
 				opt_use_index=NO set_port_name "${pkg_portdir}"
 			fi
 		}; then
-			set_pkg_pkgdir "${port_name}"
 			status=done
+			set_pkg_pkgdir "${port_name}"
 		else
 			log="install error"
-			restore_package "${old_package}" || return 1
+			restore_package "${old_package}" || {
+				warn "Failed to restore the old version," \
+				"please reinstall '${old_package}' manually."
+				return 1
+			}
 		fi
 	else
 		log="deinstall error"
 	fi
 
-	restore_required_by "${old_required_by}" "${pkg_pkgdir}/+REQUIRED_BY" || return 1
-	process_old_package "${old_package}" || :
-	clean_shlibs || warn "Couldn't remove '${PKGCOMPATDIR}'."
-	cleanup "${REPLACE_TMPDIR}"
+	restore_file "${old_required_by}" "${pkg_pkgdir}/+REQUIRED_BY" || \
+		warn "Failed to restore the +REQUIRED_BY file."
+	process_package "${old_package}" || \
+		warn "Failed to save the old package."
+	clean_shlibs || \
+		warn "Failed to clean the old shared libraries."
+	cleanup "${tmpdir}"
 
 	[ "${status}" = done ] || return 1
 
-#ifdef WITH_OPENBSD
-}
-#else
+#ifndef WITH_OPENBSD
 	if [ "${pkg_name}" != "${port_name}" ]; then
 		info "Updating the dependencies"
 		try update_dependencies "${pkg_name}" "${port_name}" || return 1
 	fi
+#endif
 }
 
 perform_version() {
@@ -966,6 +1083,7 @@ perform_version() {
 	fi
 }
 
+#ifndef WITH_OPENBSD
 update_dependencies() {
 	local oldname newname required_by dep
 
@@ -1036,7 +1154,6 @@ preserve_shlibs() {
 		info "Preserving the shared libraries"
 		make_dirs "${PKGCOMPATDIR}" || return 1
 		try cp -f ${preserved_files} "${PKGCOMPATDIR}" || return 1
-		try ldconfig -m "${PKGCOMPATDIR}" || return 1
 	fi
 }
 
@@ -1061,7 +1178,6 @@ clean_shlibs() {
 
 	if [ -n "${delete_files}" ]; then
 		try rm -f ${delete_files} || return 1
-		try ldconfig -m "${PKGCOMPATDIR}" || return 1
 	fi
 }
 
@@ -1211,17 +1327,15 @@ main() {
 		set -- $(pkg_sort "$@")
 	fi
 
+	unset do_replace
 	if [ ${opt_debug} -eq 1 ]; then
 		while [ $# -gt 0 ]; do
 			echo "${0##*/} $1"
 			shift
 		done
 	elif [ $# -gt 0 ]; then
-		do_replace=
 		create_log "${opt_log_output}" || usage 1
-
 		perform ${1+"$@"}
-
 		show_log
 	fi
 
